from .check_definitions import check_definitions
from .check_tasks import check_tasks
from .check_types import check_types


def check_input(argd):
    """Check whether the user input is valid

    Args:
        argd (dict): user input dict
    Returns:
        same as input if no errors found
        otherwise exit with a message.
    """
    # check whether the required fields exist
    required = []
    for k in required:
        if k not in argd:
            print(
                "Error hint: missing required field: \"{}\"".format(k))
            exit()
        else:
            pass

    # check whether the formats are correct
    check_definitions(argd['rule'], 'rule')
    check_tasks(argd['task'])
    check_definitions(argd['def'], 'def')
    check_types(argd['include'], 'include', [list, str])
    check_types(argd['subninja'], 'subninja', [list, str])
    check_types(argd['phony'], 'phony', [list, dict])
    check_types(argd['default'], 'default', [list, str])

    return argd



from .ninja_io import ninja_io


def ninja_rule(rules):
    """Return a list of Ninja rules

    Args:
        rules (list): a list of dictionaries of rules
    Returns:
        a list, each item is a string like
        rule <rule name>
            command = ...
    """
    output = []
    for d in rules:
        for k in sorted(d.keys()):
            v = d[k]
            if k in ['include', 'subninja']:
                output.append(
                    "{}: {}".format(k, ninja_io(v))
                    )
            else:
                output.append(
                    "rule {}\n  command = {}".format(k, v)
                )
    return "\n\n".join(output)



import sys
import os
from .core import ninja_generate
from .io.read import read_config_files
from .tk import ioTK


def main(f_out, args):
    if os.path.splitext(f_out)[1] == ".yaml":
        args = [f_out] + args
        f_out = "build.ninja"
    else:
        pass

    ioTK.save_text(
            f_out,
            ninja_generate(read_config_files(args))
        )
    print("output =", f_out)



from .ninja_def import ninja_def
from .ninja_io import ninja_io
from .ninja_build_dep import ninja_build_dep


def ninja_build(tasks):
    """Return the Ninja build section

    Args:
        tasks (list): a list of task dictionaries
    Returns:
        a string
    """
    output = []
    for t in tasks:
        item = "build {0}: {1} {2}{3}{4}".format(
                    ninja_io(t['out']),
                    t['rule'],
                    ninja_io(t['in']),
                    ninja_build_dep(t, "dep|"),
                    ninja_build_dep(t, "dep||")
                )
        if 'def' in t:
            output.append("{}\n{}".format(
                    item,
                    ninja_def(t['def'], "  "))
                )
        else:
            output.append(item)
    return "\n\n".join(output)



"""
Compose two function with single argument into one function
"""


def compose2(f1, f2):
    """Compose two function with single argument into one function

    Args:
        f1 (object): function 1
        f2 (object): function 2

    Returns:
        another function f2(f1(_))
    """
    return lambda x: f2(f1(x))



from .ninja_io import ninja_io


def ninja_def(defs, indent=""):
    """Return a list of user defined constants

    If the value of the const definition is a list,
    then it will be concatenated.

    Args:
        defs (list): a list of dictionaries of constant definitions
        indent (str): indentation for each definition

    Returns:
        a string
    """
    output = []
    for d in defs:
        for k in sorted(d.keys()):
            v = d[k]
            if k in ['include', 'subninja']:
                output.append(
                    "\n{}: {}\n".format(k, ninja_io(v))
                    )
            else:
                output.append(
                        "{}{} = {}".format(indent, k, ninja_io(v))
                    )
    return "\n".join(output)



"""
Take many functions and compose
"""
from typing import List, Callable, Optional
from .reduce import reduce
from .compose2 import compose2


def compose(functions):
    # type: (List[Callable]) -> Callable
    """Compose functions

    Args:
        functions (list): a list of functions

    Returns:
        a new function with combined effects from all inputs
    """
    return reduce(compose2, functions, lambda a: a)



def ninja_key_value_pairs(key, values):
    """Add include/subninja fields
    Args:
        key (str): such as "include", "subninja"
        values (list): a list of strings
    Returns:
        a string
    """
    return "\n\n".join(
            map(
                lambda x: "{}: {}".format(key, x),
                values
                )
        )



from .ninja_def import ninja_def
from .ninja_rule import ninja_rule
from .ninja_build import ninja_build
from .ninja_key_value_pairs import ninja_key_value_pairs
from .ninja_phony import ninja_phony


def ninja_generate(argd):
    """Return the transformed Ninja build file content

    Args:
        argd (dict): user configuration dictionary
    Returns:
        a string (content of the Ninja build file)
    """
    return "\n\n\n".join(
        filter(
                lambda x: x != "",
                [
                    ninja_def(argd['def']),
                    ninja_rule(argd['rule']),
                    ninja_build(argd['task']),
                    ninja_phony(argd['phony']),
                    ninja_key_value_pairs('include', argd['include']),
                    ninja_key_value_pairs('subninja', argd['subninja']),
                    ninja_key_value_pairs("default", argd['default'])
                ]
            )
        )



def read_all(*files):
    """Read file(s) and return the content

    Args:
        files (str): variable number of file names
    Returns:
        combined content as a string
    """
    output = []
    for f in files:
        with open(f, 'r') as IN:
            output.append(IN.read())
    return "\n".join(output)



def check_definitions(defs, field):
    """check const/rules definitions
    """
    if not isinstance(defs, list):
        print("Error hint: \"{}\" section must be a list".format(field))
        print("Your \"{}\" section has type: {}".format(field, type(defs)))
        exit()
    else:
        pass

    for d in defs:
        if not isinstance(d, dict):
            msg = ""
            msg += "Error hint: each \"{}\" ".format(field)
            msg += "definition must be a \"dict\"\n"
            msg += "Error hint: your {} definition:\n".format(field)
            msg += "\t{}\nhas type: {}".format(d, type(d))
            print(msg)
            exit()
    return True



def check_types(obj, name, types, n=0):
    """check whether an object is a list
    Args:
        obj (object): an unknown object
        name (str): name of the object
        types (list): a hierarchy of types
    """
    if len(types) == 0:
        return True
    else:
        if not isinstance(obj, types[0]):
            msg = "Error hint: field {}".format(name)
            msg += " must be of type {}".format(str(types[0]))
            msg += "The received type is {}".format(type(obj))
            print(msg)
            exit()

        # check each item has the required type
        if types[0] == list:
            next_level = "{}: level {}".format(name, n)

            def aux(obj):
                return check_types(obj, next_level, types[1:], n+1)
            return list(map(aux, obj))



"""
Ninjag: Ninja build file generator

Usage: ninjag output.ninja in1.yaml in2.yaml ...
Author: Steven(Yuhang) Wang
License: MIT/X11
"""
import sys
from .main import main
from .show_usage import show_usage


def cli():
    if len(sys.argv) < 2:
        show_usage()
        print("Error hint: need at least one command line arguments\n")
        exit()
    else:
        main(sys.argv[1], sys.argv[2:])



import re


def ninja_build_dep(task, key):
    """Add Ninja build dependency

    note: task['dep'] can be a string or list
    Args:
        task (dict): a specific task
        key (str): either 'dep|' or 'dep||'
        sep (str): the separator, either "|" or "||"
            "|" means implicit dependency
            "||" means order-only dependency
            (ninja-build.org/manual.html#ref_dependencies)
    Returns:
        an empty string if task['dep']
        doesn't exist;
        returns " | dep1 dep2" if present
    """
    # extract "|" or "||" from key
    sep = re.sub(r"\w+", "", key)

    if key in task:
        if isinstance(task[key], str):
            return " {} {}".format(sep, task[key])
        elif isinstance(task[key], list):
            return " {} {}".format(
                sep,
                " ".join(map(str, task[key]))
                )
        else:
            return ""
    else:
        return ""



from .check_definitions import check_definitions


def check_tasks(tasks):
    """Check tasks section"""
    if not isinstance(tasks, list):
        print("Error hint: \"tasks\" section must be a list")
        exit()

    for i, d in enumerate(tasks):
        for k in ["rule", "in", "out"]:
            if k not in d:
                msg = "Error hint:  task[{}] ".format(i+1)
                msg += "is missing required field: \"{}\"".format(k)
                print(msg)
                exit()
            else:
                pass

        if "const" in d:
            check_definitions(d['const'], "tasks/const")
        else:
            pass

    return True



def save_text(f_out, *text):
    """Save text to file
    Args:
        f_out (str): output file name
        text (str): variable number of strings of
            file content
    """
    with open(f_out, 'w') as OUT:
        OUT.write("\n".join(text) + "\n")



def ninja_phony(aliases):
    """Return the Ninja phony build section

    Args:
        aliases (list): a list of alias dictionaries
    Returns:
        a string
    """
    return "\n\n".join(
        map(
            lambda x: "build {}: phony {}".format(
                list(x.keys())[0],
                list(x.values())[0]
                ),
            aliases
            )
    )



"""
Read all the user configuration files
and return a combined dictionary
"""
import yaml
from ...tk import fnTK, dictTK, ioTK
from .check_input import check_input
from .default_config import default_config


def read_config_files(files):
    """Read config files

    Args:
        files (str): a list of file names
    Returns:
        a dict
    """
    fn = fnTK.compose([ioTK.read_all, yaml.load])
    return check_input(
        dictTK.merge([default_config()] + list(map(fn, files))))
